"""Biosample parser

Implemented as an iterator


>>> import omicidx.biosample as b
>>> for bios in b.BioSampleParser(gzip.open('biosample_set.xml.gz', 'rb')):
>>>     print(bios.json())
>>>     print(bios.dict())
"""

import os.path
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element
import gzip
import json
import typing
import logging
import click
import subprocess
import re
import pydantic
import datetime


class IdRecs(pydantic.BaseModel):
    db: typing.Optional[str]
    label: typing.Optional[str]
    id: typing.Optional[str]


class AttrRecs(pydantic.BaseModel):
    attribute_name: typing.Optional[str]
    display_name: typing.Optional[str]
    harmonized_name: typing.Optional[str]
    value: typing.Optional[str]
    unit: typing.Optional[str]


class BioSample(pydantic.BaseModel):
    accession: str
    id: str
    title: typing.Optional[str]
    description: typing.Optional[str]
    taxonomy_name: str
    taxon_id: int
    attribute_recs: list[AttrRecs]
    attributes: list[str]
    model: typing.Optional[str]
    id_recs: list[IdRecs]
    ids: list[str]
    sra_sample: typing.Optional[str]
    dbgap: typing.Optional[str]
    gsm: typing.Optional[str]
    publication_date: typing.Optional[datetime.datetime]
    last_update: typing.Optional[datetime.datetime]
    submission_date: typing.Optional[datetime.datetime]
    access: typing.Optional[str]


class SimplePublication(pydantic.BaseModel):
    db: typing.Optional[str]
    id: typing.Optional[str]
    pubdate: typing.Optional[datetime.datetime]


class ExternalLink(pydantic.BaseModel):
    url: typing.Optional[str]
    label: typing.Optional[str]
    category: typing.Optional[str]


class BioProject(pydantic.BaseModel):
    data_types: list[str] = []
    description: typing.Optional[str]
    accession: str
    name: typing.Optional[str]
    publications: list[SimplePublication] = []
    title: typing.Optional[str]
    external_links: list[ExternalLink] = []


class BioSampleParser(object):
    """Parse a BioSample xml file.

    This is a generator that yields dict records.
    If you want to validate the records, set validate_with_schema to True
    and the generator will pass the records through pydantic models.
    """

    def __init__(self, fh: typing.IO, validate_with_schema: bool = False):
        """Initialize a new parser

        Args:
            fh (typing.IO): An open file-like object containing Biosample XML records.
        """
        self.fhandle = fh
        self.context = ET.iterparse(self.fhandle, events=("end",))
        self.validate_with_schema = validate_with_schema
        event, self.root = next(self.context)

    def __iter__(self):
        return self

    def __next__(self):
        for event, elem in self.context:
            if elem.tag == "BioSample":
                bios = dict()
                bios["is_reference"] = None
                for k, v in elem.items():
                    bios[k] = v
                bios["id_recs"] = []
                bios["ids"] = []
                bios["sra_sample"] = None
                bios["dbgap"] = None
                bios["gsm"] = None
                for id in elem.iterfind(".//Id"):
                    idrec = {
                        "db": id.get("db"),
                        "label": id.get("db_label"),
                        "id": id.text,
                    }
                    bios["ids"].append(idrec["id"])
                    bios["id_recs"].append(idrec)
                    # add xref fields for SRA, dbGaP, and GEO

                    if id.get("db") == "SRA":
                        bios["sra_sample"] = id.text
                    if id.get("db") == "dbGaP":
                        bios["dbgap"] = id.text
                    if id.get("db") == "GEO":
                        bios["gsm"] = id.text
                bios["title"] = elem.findtext(".//Description/Title")
                bios["description"] = elem.findtext(".//Description/Comment/Paragraph")
                organism = elem.find(".//Organism")
                bios["taxonomy_name"] = organism.get("taxonomy_name")
                bios["taxon_id"] = int(organism.get("taxonomy_id"))
                bios["attribute_recs"] = []
                bios["attributes"] = []
                for attribute in elem.findall(".//Attribute"):
                    attrec = attribute.attrib
                    attrec["value"] = attribute.text
                    bios["attribute_recs"].append(attrec)
                    try:
                        bios["attributes"].append(attrec["harmonized_name"])
                    except:
                        bios["attributes"].append(attrec["attribute_name"])

                bios["model"] = elem.findtext(".//Model")
                # print(json.dumps(bios))
                # res = es.index(index="bioes", doc_type='biosample', id=bios['id'], body=bios)
                elem.clear()

                # allow this behavior to be turned off
                # as performance is better without validation
                if self.validate_with_schema:
                    return BioSample(**bios)
                else:
                    return bios
        raise StopIteration


def parse_bioproject_xml_element(element: Element) -> dict:
    """Parse a BioProject xml element

    Args:
        element (Element): An lxml.etree.Element

    Returns:
        dict: A BioProject dict.
    """
    projtop: Element = element.find("./Project")
    d2 = {}
    d2["title"] = projtop.findtext("./Project/ProjectDescr/Title")
    d2["description"] = projtop.findtext("./Project/ProjectDescr/Description")
    d2["name"] = projtop.findtext("./Project/ProjectDescr/Name")
    archive_id = projtop.find("./Project/ProjectID/ArchiveID")
    d2["accession"] = archive_id.attrib["accession"]
    pubs = []
    for pub in projtop.findall(".//Publication"):
        pubs.append(
            {
                "pubdate": pub.get("date", None),
                "id": pub.get("id", None),
                "db": re.sub("^e", "", pub.findtext("./DbType").replace("^e", "", 1)),
            }
        )
    d2["publications"] = pubs
    ext_links = []
    for link in projtop.findall(".//ExternalLink"):
        ext_links.append(
            {
                "category": link.get("category", None),
                "label": link.get("label", None),
                "url": link.findtext("./URL"),
            }
        )
    data_types = []
    for datatype in projtop.findall(".//ProjectDataTypeSet"):
        data_types.append(datatype.findtext("./DataType"))
    d2["data_types"] = data_types
    d2["external_links"] = ext_links
    return d2


class BioProjectParser(typing.Iterable):
    """Parse a BioProject xml file.

    If you want to validate the records, set validate_with_schema to True
    and the generator will pass the records through pydantic models, although
    this will be slower (by about 5x).

    The return values are simple dicts.
    """

    def __init__(self, fh: typing.IO, validate_with_schema: bool = False):
        """Initialize a BioProjectParser

        Args:
            fh (typing.IO): An open file-like object containing BioProject records.
        """
        self.fhandle = fh
        self.context = ET.iterparse(self.fhandle, events=("end",))
        self.validate_with_schema = validate_with_schema
        event, self.root = next(self.context)

    def __iter__(self):
        return self

    def __next__(self):
        for event, elem in self.context:
            if elem.tag == "Package":
                results = parse_bioproject_xml_element(elem)
                elem.clear()
                if self.validate_with_schema:
                    return BioProject(**results).dict()
                else:
                    return results

        raise StopIteration



from pydantic import BaseModel, ConfigDict
from datetime import datetime
from typing import List, Dict, Optional


class Attribute(BaseModel):
    tag: Optional[str] = None
    value: Optional[str] = None


class Xref(BaseModel):
    db: Optional[str] = None
    id: Optional[str] = None


class Identifier(BaseModel):
    namespace: Optional[str] = None
    id: Optional[str] = None


class FileAlternative(BaseModel):
    url: Optional[str] = None
    free_egress: Optional[str] = None
    access_type: Optional[str] = None
    org: Optional[str] = None


class FileSet(BaseModel):
    cluster: str = "public"
    filename: Optional[str] = None
    url: Optional[str] = None
    size: int = 0
    date: datetime = None
    md5: Optional[str] = None
    sratoolkit: str = "1"
    alternatives: List[FileAlternative]


class BaseQualityCount(BaseModel):
    quality: int = 0
    count: int = 0


class BaseQualities(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    pass


class TaxCountEntry(BaseModel):
    rank: Optional[str] = None
    name: Optional[str] = None
    parent: Optional[int] = None
    total_count: int = 0
    self_count: int = 0
    tax_id: int


class TaxCountAnalysis(BaseModel):
    nspot_analyze: Optional[int] = None
    total_spots: Optional[int] = None
    mapped_spots: Optional[int] = None
    tax_counts: List[TaxCountEntry] = None


class RunRead(BaseModel):
    index: int
    count: int
    mean_length: float = 0.0
    sd_length: float = 0.0


class BaseCounts(BaseModel):  # (List[Dict[str, int]]):
    pass


class LiveList(BaseModel):
    lastupdate: datetime = None
    published: datetime = None
    received: datetime = None
    status: str = "live"
    insdc: bool = True


class SraRun(LiveList, BaseModel):
    alias: Optional[str] = None
    run_date: datetime = None
    run_center: Optional[str] = None
    center_name: Optional[str] = None
    accession: str
    total_spots: int = 0
    total_bases: int = 0
    size: int = 0
    load_done: bool = True
    published: datetime = None
    is_public: bool = True
    cluster_name: str = "public"
    static_data_available: str = "1"
    avg_length: float = 0.0
    experiment_accession: str
    attributes: List[Attribute] = None
    files: List[FileSet] = None
    qualities: BaseQualities = None
    base_counts: BaseCounts = None
    reads: List[RunRead] = None
    tax_analysis: TaxCountAnalysis = None


class SraStudy(LiveList, BaseModel):
    abstract: Optional[str] = None
    BioProject: Optional[str] = None
    Geo: Optional[str] = None
    accession: str
    alias: Optional[str] = None
    center_name: Optional[str] = None
    broker_name: Optional[str] = None
    description: Optional[str] = None
    study_type: Optional[str] = None
    title: Optional[str] = None
    identifiers: List[Identifier] = None
    attributes: List[Attribute] = None
    pubmed_ids: List[int] = None


class SraExperiment(LiveList, BaseModel):
    accession: str
    attributes: List[Attribute] = None
    alias: Optional[str] = None
    center_name: Optional[str] = None
    design: Optional[str] = None
    description: Optional[str] = None
    identifiers: List[Identifier] = None
    instrument_model: Optional[str] = None
    library_name: Optional[str] = None
    library_construction_protocol: Optional[str] = None
    library_layout_orientation: Optional[str] = None
    library_layout_length: float = None
    library_layout_sdev: float = None
    library_strategy: Optional[str] = None
    library_source: Optional[str] = None
    library_selection: Optional[str] = None
    library_layout: Optional[str] = None
    xrefs: List[Xref] = None
    platform: Optional[str] = None
    sample_accession: Optional[str] = None
    study_accession: Optional[str] = None
    title: Optional[str] = None


class SraSample(LiveList, BaseModel):
    accession: str
    geo: Optional[str] = None
    BioSample: Optional[str] = None
    title: Optional[str] = None
    alias: Optional[str] = None
    organism: Optional[str] = None
    taxon_id: Optional[int] = None
    description: Optional[str] = None
    identifiers: List[Identifier] = None
    attributes: List[Attribute] = None
    xrefs: List[Xref] = None


class FullSraRun(SraRun):
    experiment: Optional[SraExperiment] = None
    sample: Optional[SraSample] = None
    study: Optional[SraStudy] = None



import omicidx.geo.parser as op
import omicidx.biosample
import click
import json
import sys
from datetime import date


@click.group()
def cli():
    pass


@cli.group("biosample")
def biosample():
    pass


@biosample.command("downloadxml")
def biosample_download():
    """Download the biosample_set.xml.gz file"""
    omicidx.biosample.download_biosample()


@biosample.command("xml2json")
@click.option(
    "--infile",
    help="compose file to work with",
    type=click.File("r"),
    default=sys.stdin,
)
def biosample_to_json(infile: click.File):
    """Convert the biosample xml file to json"""
    bp = omicidx.biosample.BioSampleParser(infile)
    print(infile.name)
    for rec in bp:
        print(rec.as_json())


@cli.command()
@click.option("--gse", help="A single GSE accession")
def gse_to_json(gse):
    print(op.gse_to_json(gse))


@cli.command()
@click.argument("etype", type=click.Choice(["GSE", "GPL", "GSM"]), default="GSE")
@click.option(
    "--date-start", type=click.DateTime(formats=["%Y-%m-%d"]), default=str(date.today())
)
@click.option(
    "--date-end", type=click.DateTime(formats=["%Y-%m-%d"]), default=str(date.today())
)
def bulk_gse_to_json(date_start: date, date_end: date, etype: str = "GSE"):
    date_formatter = lambda x: x.strftime("%Y/%m/%d")
    term = f"{date_formatter(date_start)}:{date_formatter(date_end)}[PDAT] AND {etype}[ETYP]"
    op.bulk_gse_to_json(term, sys.stdout)


@cli.command()
@click.option("--term", help='something like "2019/01/28:2019/02/28[PDAT]"')
@click.option("--outfile", help="output file name, default is stdout")
def gse_accessions(term, outfile="/dev/stdout"):
    op.gse_accessions(term, outfile)


if __name__ == "__main__":
    cli()



import pydantic
from pydantic import BaseModel, constr
import json
from datetime import datetime, date
from typing import List, Dict, Any, Optional


class GEOBase(BaseModel):
    title: str
    status: str
    submission_date: Optional[date] = None
    last_update_date: Optional[date] = None


class GEOName(BaseModel):
    first: Optional[str] = None
    middle: Optional[str] = None
    last: Optional[str] = None


class GEOContact(BaseModel):
    city: Optional[str] = None
    name: GEOName = None
    email: Optional[str] = None
    state: Optional[str] = None
    address: Optional[str] = None
    department: Optional[str] = None
    country: Optional[str] = None
    web_link: Optional[str] = None
    institute: Optional[str] = None
    zip_postal_code: Optional[str] = None
    phone: Optional[str] = None


class GEOPlatform(GEOBase):
    accession: str  # constr(pattern="GPL[0-9]+")
    status: str
    _entity: str = "GPL"
    contact: GEOContact = None
    summary: Optional[str] = None
    organism: Optional[str] = None
    sample_id: Optional[List[str]]  # List[constr(pattern="GSM[0-9]+")] = []
    series_id: Optional[List[str]]  # List[constr(pattern="GSE[0-9]+")] = []
    technology: Optional[str] = None
    description: Optional[str] = None
    distribution: Optional[str] = None
    manufacturer: List[str] = []
    data_row_count: Optional[int] = None
    contributor: List[GEOName] = []
    relation: List[str] = []
    manufacture_protocol: Optional[str] = None


class GEOCharacteristic(BaseModel):
    tag: str
    value: Optional[str] = None  # there are apparently some of these


class GEOChannel(BaseModel):
    label: Optional[str] = None
    taxid: List[int] = []
    molecule: Optional[str] = None
    organism: Optional[str] = None
    source_name: Optional[str] = None
    label_protocol: Optional[str] = None
    growth_protocol: Optional[str] = None
    extract_protocol: Optional[str] = None
    treatment_protocol: Optional[str] = None
    characteristics: List[GEOCharacteristic] = []


class GEOSample(GEOBase):
    type: str
    anchor: Optional[str] = None
    _entity: None
    contact: GEOContact = None
    description: Optional[str] = None
    accession: constr(pattern="GSM[0-9]+")
    biosample: Optional[constr(pattern="SAM[A-Z]+[0-9]+")] = None
    tag_count: int = None
    tag_length: float = None
    platform_id: constr(pattern="GPL[0-9]+")
    hyb_protocol: Optional[str] = None
    channel_count: int = 0
    scan_protocol: Optional[str] = None
    data_row_count: int = 0
    library_source: Optional[str] = None
    overall_design: Optional[str] = None
    sra_experiment: Optional[constr(pattern="[DES]RX[0-9]+")] = None
    data_processing: Optional[str] = None
    supplemental_files: List[str] = []
    channels: List[GEOChannel] = []
    contributor: List[GEOName] = []


class GEOSeries(GEOBase):
    accession: constr(pattern="GSE[0-9]+")
    subseries: List[constr(pattern="GSE[0-9]+")] = []
    bioprojects: List[constr(pattern="PRJ[A-Z]+[0-9]+")] = []
    sra_studies: List[constr(pattern="[ESD]RP[0-9]+")] = []
    _entity: str = "GSE"
    contact: GEOContact = None
    type: List[str] = []
    summary: Optional[str] = None
    relation: List[str] = []
    pubmed_id: List[int] = []
    sample_id: List[constr(pattern="GSM[0-9]+")] = []
    sample_taxid: List[int] = []
    sample_organism: List[str] = []
    platform_id: List[constr(pattern="GPL[0-9]+")] = []
    platform_taxid: List[int] = []
    platform_organism: List[str] = []
    data_processing: Optional[str] = None
    description: Optional[str] = None
    supplemental_files: List[str] = []
    overall_design: Optional[str] = None
    contributor: List[GEOName] = []



import click
import omicidx.biosample
from sd_cloud_utils.aws.sqs import SQS
import omicidx.sra.parser as sp
import logging
import json
import datetime
import asyncio
import sys

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('omicidx-cli')

@click.group('biosample')
def biosample():
    pass

@biosample.command('downloadxml')
def biosample_download():
    """Download the biosample_set.xml.gz file"""
    omicidx.biosample.download_biosample()

@biosample.command('xml2json')
@click.option('--infile', 
              help='compose file to work with',
              type=click.File('r'),
              default=sys.stdin)
def biosample_to_json(xmlfilehandle):
    """Convert the biosample xml file to json"""
    bp = omicidx.biosample.BioSampleParser(xmlfilehandle)
    for rec in bp:
        print(rec.as_json())

@click.group('sra')
#@click.pass_context
def sra():
    pass

@sra.command('download')
def sra_download():
    print('Downloading files')

@sra.command('entity2json')
@click.argument("entity")
def sra_entity_to_json(entity):
    print('convert one entity')

@sra.command('entities2json')
def sra_entities2json():
    print('convert everything at once')

@sra.command("load2postgres")
def sra_load2postgres():
    print('load and transform in postgresql')

@sra.command("load2bigquery")
def sra_load2bigquery():
    print("Load to bigquery")

@click.group(help='get SRA metadata as json')
@click.option('--config',
              help="config file")
def json_utils(config=None):
    pass


def dateconverter(o):
    if isinstance(o, datetime.datetime):
        return o.__str__()

def _srr_to_json(accession):
    models = sp.models_from_runbrowser(accession)
    if(not isinstance(models, dict)):
        return None
    res = {}
    for k in models.keys():
        res[k] = models[k].dict()
    print(json.dumps(res, default = dateconverter))

@json_utils.command(help='Convert SRR to json')
@click.option('--accession',
              help = "An SRA Run accession")
def srr_to_json(accession):
    _srr_to_json(accession)
    
@click.group()
def omicidx_cli():
    pass

@click.group()
def admin():
    pass

@admin.command()
@click.option('--queue')
def create_queue(queue):
    sqs = SQS(queuename=queue)
    sqs.get_queue()
    return True

@admin.command()
@click.option('--queue')
@click.option('--from_date')
@click.option('--to_date')
def send_message(queue, from_date, to_date):
    n = 0
    q = SQS(queue)
    msg = []
    for rec in sp.get_accession_list(
            from_date = from_date,
            to_date = to_date,
            type = 'RUN'
    ):
        msg.append(q.make_message(body=rec['Accession'], id=str(n)))
        n+=1
        if(n % 10 == 0):
            q.send_messages(msg)
            msg = []
            logger.info('sent {} messages'.format(n))
    q.send_messages(msg)
    logger.info('sent {} messages'.format(n))
    


async def produce(queue, queuename):
    while True:
        vals = SQS(queuename=queuename).receive_message()
        if(len(vals)==0):
            logger.info('nothing to retrieve')
            queue.close()
        for val in vals:
            logger.info('Message:' + val.body)
            await queue.put(val)


async def consume(queue):
    while True:
        # wait for an item from the producer
        val = await queue.get()
        
        # process the item
        models = sp.models_from_runbrowser(val.body)
        if(isinstance(models, dict)):
            res = {}
            for k in models.keys():
                res[k] = models[k].dict()
            val.delete()
            queue.task_done()
            print(json.dumps(res, default = dateconverter))
        else:
            # Notify the queue that the item has been processed
            val.delete()
            queue.task_done()


async def run(queuename):
    queue = asyncio.Queue(maxsize=10)
    # schedule the consumer
    consumer = asyncio.ensure_future(consume(queue))
    # run the producer and wait for completion
    await produce(queue, queuename)
    # wait until the consumer has processed all items
    await queue.join()
    # the consumer is still awaiting for an item, cancel it
    consumer.cancel()



@admin.command(help='write out livelist results to stdout')
@click.option('--from_date', default = "2001-01-01",
              help = "records from date (default '2001-01-01')")
@click.option('--to_date',default=None, type=str)
@click.option('--entity',
              help = 'type of record to return',
              type = click.Choice(['RUN', 'EXPERIMENT', 'SAMPLE', 'STUDY']),
              default="RUN")
@click.option('--count',
              help="records to retrieve in each call (default 1000)",
              default = 1000,
              type=int)
@click.option('--offset',
              help="start at offset (default 0)",
              default = 0,
              type=int)
@click.option('--header', flag_value='header',
              help="include header (default False)",
              default = False)
def livelist(from_date, to_date, entity, count, offset, header):
    ll = sp.LiveList(from_date = from_date,
                     to_date = to_date, offset = offset,
                     count = count, entity = entity)
    if(header):
        print("\t".join(row))
    for row in ll:
        print('\t'.join(row.values()))
    
                     

            
@admin.command()
@click.option('--queue')
def process_accessions(queue):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run(queue))
    loop.close()



omicidx_cli.add_command(biosample)
omicidx_cli.add_command(sra)
omicidx_cli.add_command(json_utils)
omicidx_cli.add_command(admin)



if __name__ == '__main__':
    omicidx_cli()
